"""API for working with a Nvim Buffer."""

from __future__ import annotations

from typing import (Any, Iterator, List, Optional, TYPE_CHECKING, Tuple, Union, cast,
                    overload)

from pynvim.api.common import Remote
from pynvim.compat import check_async

if TYPE_CHECKING:
    from pynvim.api import Nvim


__all__ = ('Buffer',)


@overload
def adjust_index(idx: int, default: Optional[int] = None) -> int:
    ...


@overload
def adjust_index(idx: Optional[int], default: int) -> int:
    ...


@overload
def adjust_index(idx: Optional[int], default: Optional[int] = None) -> Optional[int]:
    ...


def adjust_index(idx: Optional[int], default: Optional[int] = None) -> Optional[int]:
    """Convert from python indexing convention to nvim indexing convention."""
    if idx is None:
        return default
    elif idx < 0:
        return idx - 1
    else:
        return idx


class Buffer(Remote):

    """A remote Nvim buffer."""

    _api_prefix = "nvim_buf_"
    _session: "Nvim"

    def __init__(self, session: Nvim, code_data: Tuple[int, Any]):
        """Initialize from Nvim and code_data immutable object."""
        super().__init__(session, code_data)

    def __len__(self) -> int:
        """Return the number of lines contained in a Buffer."""
        return self.request('nvim_buf_line_count')

    @overload
    def __getitem__(self, idx: int) -> str:  # noqa: D105
        ...

    @overload
    def __getitem__(self, idx: slice) -> List[str]:  # noqa: D105
        ...

    def __getitem__(self, idx: Union[int, slice]) -> Union[str, List[str]]:
        """Get a buffer line or slice by integer index.

        Indexes may be negative to specify positions from the end of the
        buffer. For example, -1 is the last line, -2 is the line before that
        and so on.

        When retrieving slices, omitting indexes(eg: `buffer[:]`) will bring
        the whole buffer.
        """
        if not isinstance(idx, slice):
            i = adjust_index(idx)
            return self.request('nvim_buf_get_lines', i, i + 1, True)[0]
        start = adjust_index(idx.start, 0)
        end = adjust_index(idx.stop, -1)
        return self.request('nvim_buf_get_lines', start, end, False)

    @overload
    def __setitem__(self, idx: int, item: Optional[str]) -> None:  # noqa: D105
        ...

    @overload
    def __setitem__(  # noqa: D105
        self, idx: slice, item: Optional[Union[List[str], str]]
    ) -> None:
        ...

    def __setitem__(
        self, idx: Union[int, slice], item: Union[None, str, List[str]]
    ) -> None:
        """Replace a buffer line or slice by integer index.

        Like with `__getitem__`, indexes may be negative.

        When replacing slices, omitting indexes(eg: `buffer[:]`) will replace
        the whole buffer.
        """
        if not isinstance(idx, slice):
            assert not isinstance(item, list)
            i = adjust_index(idx)
            lines = [item] if item is not None else []
            return self.request('nvim_buf_set_lines', i, i + 1, True, lines)
        if item is None:
            lines = []
        elif isinstance(item, str):
            lines = [item]
        else:
            lines = item
        start = adjust_index(idx.start, 0)
        end = adjust_index(idx.stop, -1)
        return self.request('nvim_buf_set_lines', start, end, False, lines)

    def __iter__(self) -> Iterator[str]:
        """Iterate lines of a buffer.

        This will retrieve all lines locally before iteration starts. This
        approach is used because for most cases, the gain is much greater by
        minimizing the number of API calls by transferring all data needed to
        work.
        """
        lines = self[:]
        for line in lines:
            yield line

    def __delitem__(self, idx: Union[int, slice]) -> None:
        """Delete line or slice of lines from the buffer.

        This is the same as __setitem__(idx, [])
        """
        self.__setitem__(idx, None)

    def __ne__(self, other: Any) -> bool:
        """Test inequality of Buffers.

        Necessary for Python 2 compatibility.
        """
        return not self.__eq__(other)

    def append(
        self, lines: Union[str, bytes, List[Union[str, bytes]]], index: int = -1
    ) -> None:
        """Append a string or list of lines to the buffer."""
        if isinstance(lines, (str, bytes)):
            lines = [lines]
        return self.request('nvim_buf_set_lines', index, index, True, lines)

    def mark(self, name: str) -> Tuple[int, int]:
        """Return (row, col) tuple for a named mark."""
        return cast(Tuple[int, int], tuple(self.request('nvim_buf_get_mark', name)))

    def range(self, start: int, end: int) -> Range:
        """Return a `Range` object, which represents part of the Buffer."""
        return Range(self, start, end)

    def add_highlight(
        self,
        hl_group: str,
        line: int,
        col_start: int = 0,
        col_end: int = -1,
        src_id: int = -1,
        async_: Optional[bool] = None,
        **kwargs: Any
    ) -> int:
        """Add a highlight to the buffer."""
        async_ = check_async(async_, kwargs, src_id != 0)
        return self.request(
            "nvim_buf_add_highlight",
            src_id,
            hl_group,
            line,
            col_start,
            col_end,
            async_=async_,
        )

    def clear_highlight(
        self,
        src_id: int,
        line_start: int = 0,
        line_end: int = -1,
        async_: Optional[bool] = None,
        **kwargs: Any
    ) -> None:
        """Clear highlights from the buffer."""
        async_ = check_async(async_, kwargs, True)
        self.request(
            "nvim_buf_clear_highlight", src_id, line_start, line_end, async_=async_
        )

    def update_highlights(
        self,
        src_id: int,
        hls: List[Union[Tuple[str, int], Tuple[str, int, int, int]]],
        clear_start: Optional[int] = None,
        clear_end: int = -1,
        clear: bool = False,
        async_: bool = True,
    ) -> None:
        """Add or update highlights in batch to avoid unnecessary redraws.

        A `src_id` must have been allocated prior to use of this function. Use
        for instance `nvim.new_highlight_source()` to get a src_id for your
        plugin.

        `hls` should be a list of highlight items. Each item should be a list
        or tuple on the form `("GroupName", linenr, col_start, col_end)` or
        `("GroupName", linenr)` to highlight an entire line.

        By default existing highlights are preserved. Specify a line range with
        clear_start and clear_end to replace highlights in this range. As a
        shorthand, use clear=True to clear the entire buffer before adding the
        new highlights.
        """
        if clear and clear_start is None:
            clear_start = 0
        lua = self._session._get_lua_private()
        lua.update_highlights(self, src_id, hls, clear_start, clear_end, async_=async_)

    @property
    def name(self) -> str:
        """Get the buffer name."""
        return self.request('nvim_buf_get_name')

    @name.setter
    def name(self, value: str) -> None:
        """Set the buffer name. BufFilePre/BufFilePost are triggered."""
        return self.request('nvim_buf_set_name', value)

    @property
    def valid(self) -> bool:
        """Return True if the buffer still exists."""
        return self.request('nvim_buf_is_valid')

    @property
    def loaded(self) -> bool:
        """Return True if the buffer is valid and loaded."""
        return self.request('nvim_buf_is_loaded')

    @property
    def number(self) -> int:
        """Get the buffer number."""
        return self.handle


class Range:

    def __init__(self, buffer: Buffer, start: int, end: int):
        self._buffer = buffer
        self.start = start - 1
        self.end = end - 1

    def __len__(self) -> int:
        return self.end - self.start + 1

    @overload
    def __getitem__(self, idx: int) -> str:
        ...

    @overload
    def __getitem__(self, idx: slice) -> List[str]:
        ...

    def __getitem__(self, idx: Union[int, slice]) -> Union[str, List[str]]:
        if not isinstance(idx, slice):
            return self._buffer[self._normalize_index(idx)]
        start = self._normalize_index(idx.start)
        end = self._normalize_index(idx.stop)
        if start is None:
            start = self.start
        if end is None:
            end = self.end + 1
        return self._buffer[start:end]

    @overload
    def __setitem__(self, idx: int, lines: Optional[str]) -> None:
        ...

    @overload
    def __setitem__(self, idx: slice, lines: Optional[List[str]]) -> None:
        ...

    def __setitem__(
        self, idx: Union[int, slice], lines: Union[None, str, List[str]]
    ) -> None:
        if not isinstance(idx, slice):
            assert not isinstance(lines, list)
            self._buffer[self._normalize_index(idx)] = lines
            return
        start = self._normalize_index(idx.start)
        end = self._normalize_index(idx.stop)
        if start is None:
            start = self.start
        if end is None:
            end = self.end
        self._buffer[start:end + 1] = lines

    def __iter__(self) -> Iterator[str]:
        for i in range(self.start, self.end + 1):
            yield self._buffer[i]

    def append(
        self, lines: Union[str, bytes, List[Union[str, bytes]]], i: Optional[int] = None
    ) -> None:
        i = self._normalize_index(i)
        if i is None:
            i = self.end + 1
        self._buffer.append(lines, i)

    @overload
    def _normalize_index(self, index: int) -> int:
        ...

    @overload
    def _normalize_index(self, index: None) -> None:
        ...

    def _normalize_index(self, index: Optional[int]) -> Optional[int]:
        if index is None:
            return None
        if index < 0:
            index = self.end
        else:
            index += self.start
            if index > self.end:
                index = self.end
        return index



"""Msgpack handling in the event loop pipeline."""
import logging

from msgpack import Packer, Unpacker

from pynvim.compat import unicode_errors_default
from pynvim.msgpack_rpc.event_loop.base import BaseEventLoop

logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warning,)


class MsgpackStream:
    """Two-way msgpack stream that wraps a event loop byte stream.

    This wraps the event loop interface for reading/writing bytes and
    exposes an interface for reading/writing msgpack documents.
    """

    def __init__(self, event_loop: BaseEventLoop) -> None:
        """Wrap `event_loop` on a msgpack-aware interface."""
        self.loop = event_loop
        self._packer = Packer(unicode_errors=unicode_errors_default)
        self._unpacker = Unpacker(unicode_errors=unicode_errors_default)
        self._message_cb = None

    def threadsafe_call(self, fn):
        """Wrapper around `BaseEventLoop.threadsafe_call`."""
        self.loop.threadsafe_call(fn)

    def send(self, msg):
        """Queue `msg` for sending to Nvim."""
        debug('sending %s', msg)
        self.loop.send(self._packer.pack(msg))

    def run(self, message_cb):
        """Run the event loop to receive messages from Nvim.

        While the event loop is running, `message_cb` will be called whenever
        a message has been successfully parsed from the input stream.
        """
        self._message_cb = message_cb
        self.loop.run(self._on_data)
        self._message_cb = None

    def stop(self):
        """Stop the event loop."""
        self.loop.stop()

    def close(self):
        """Close the event loop."""
        self.loop.close()

    def _on_data(self, data: bytes) -> None:
        self._unpacker.feed(data)
        while True:
            try:
                debug('waiting for message...')
                msg = next(self._unpacker)
                debug('received message: %s', msg)
                assert self._message_cb is not None
                self._message_cb(msg)  # type: ignore[unreachable]
            except StopIteration:
                debug('unpacker needs more data...')
                break



"""Msgpack-rpc subpackage.

This package implements a msgpack-rpc client. While it was designed for
handling some Nvim particularities(server->client requests for example), the
code here should work with other msgpack-rpc servers.
"""
from typing import Any, List

from pynvim.msgpack_rpc.async_session import AsyncSession
from pynvim.msgpack_rpc.event_loop import EventLoop, TTransportType
from pynvim.msgpack_rpc.msgpack_stream import MsgpackStream
from pynvim.msgpack_rpc.session import ErrorResponse, Session
from pynvim.util import get_client_info


__all__ = ('tcp_session', 'socket_session', 'stdio_session', 'child_session',
           'ErrorResponse')


def session(
    transport_type: TTransportType = 'stdio', *args: Any, **kwargs: Any
) -> Session:
    loop = EventLoop(transport_type, *args, **kwargs)
    msgpack_stream = MsgpackStream(loop)
    async_session = AsyncSession(msgpack_stream)
    session = Session(async_session)
    session.request(b'nvim_set_client_info',
                    *get_client_info('client', 'remote', {}), async_=True)
    return session


def tcp_session(address: str, port: int = 7450) -> Session:
    """Create a msgpack-rpc session from a tcp address/port."""
    return session('tcp', address, port)


def socket_session(path: str) -> Session:
    """Create a msgpack-rpc session from a unix domain socket."""
    return session('socket', path)


def stdio_session() -> Session:
    """Create a msgpack-rpc session from stdin/stdout."""
    return session('stdio')


def child_session(argv: List[str]) -> Session:
    """Create a msgpack-rpc session from a new Nvim instance."""
    return session('child', argv)



"""Decorators used by python host plugin system."""

import inspect
import logging
import sys
from typing import Any, Callable, Dict, Optional, TypeVar, Union

from pynvim.compat import unicode_errors_default

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warning,)
__all__ = ('plugin', 'rpc_export', 'command', 'autocmd', 'function',
           'encoding', 'decode', 'shutdown_hook')

T = TypeVar('T')
F = TypeVar('F', bound=Callable[..., Any])


def plugin(cls: T) -> T:
    """Tag a class as a plugin.

    This decorator is required to make the class methods discoverable by the
    plugin_load method of the host.
    """
    cls._nvim_plugin = True  # type: ignore[attr-defined]
    # the _nvim_bind attribute is set to True by default, meaning that
    # decorated functions have a bound Nvim instance as first argument.
    # For methods in a plugin-decorated class this is not required, because
    # the class initializer will already receive the nvim object.
    predicate = lambda fn: hasattr(fn, '_nvim_bind')
    for _, fn in inspect.getmembers(cls, predicate):
        fn._nvim_bind = False
    return cls


def rpc_export(rpc_method_name: str, sync: bool = False) -> Callable[[F], F]:
    """Export a function or plugin method as a msgpack-rpc request handler."""
    def dec(f: F) -> F:
        f._nvim_rpc_method_name = rpc_method_name  # type: ignore[attr-defined]
        f._nvim_rpc_sync = sync  # type: ignore[attr-defined]
        f._nvim_bind = True  # type: ignore[attr-defined]
        f._nvim_prefix_plugin_path = False  # type: ignore[attr-defined]
        return f
    return dec


def command(
    name: str,
    nargs: Union[str, int] = 0,
    complete: Optional[str] = None,
    range: Optional[Union[str, int]] = None,
    count: Optional[int] = None,
    bang: bool = False,
    register: bool = False,
    sync: bool = False,
    allow_nested: bool = False,
    eval: Optional[str] = None
) -> Callable[[F], F]:
    """Tag a function or plugin method as a Nvim command handler."""
    def dec(f: F) -> F:
        f._nvim_rpc_method_name = (  # type: ignore[attr-defined]
            'command:{}'.format(name)
        )
        f._nvim_rpc_sync = sync  # type: ignore[attr-defined]
        f._nvim_bind = True  # type: ignore[attr-defined]
        f._nvim_prefix_plugin_path = True  # type: ignore[attr-defined]

        opts: Dict[str, Any] = {}

        if range is not None:
            opts['range'] = '' if range is True else str(range)
        elif count is not None:
            opts['count'] = count

        if bang:
            opts['bang'] = ''

        if register:
            opts['register'] = ''

        if nargs:
            opts['nargs'] = nargs

        if complete:
            opts['complete'] = complete

        if eval:
            opts['eval'] = eval

        if not sync and allow_nested:
            rpc_sync: Union[bool, Literal['urgent']] = "urgent"
        else:
            rpc_sync = sync

        f._nvim_rpc_spec = {  # type: ignore[attr-defined]
            'type': 'command',
            'name': name,
            'sync': rpc_sync,
            'opts': opts
        }
        return f
    return dec


def autocmd(
    name: str,
    pattern: str = '*',
    sync: bool = False,
    allow_nested: bool = False,
    eval: Optional[str] = None
) -> Callable[[F], F]:
    """Tag a function or plugin method as a Nvim autocommand handler."""
    def dec(f: F) -> F:
        f._nvim_rpc_method_name = (  # type: ignore[attr-defined]
            'autocmd:{}:{}'.format(name, pattern)
        )
        f._nvim_rpc_sync = sync  # type: ignore[attr-defined]
        f._nvim_bind = True  # type: ignore[attr-defined]
        f._nvim_prefix_plugin_path = True  # type: ignore[attr-defined]

        opts = {
            'pattern': pattern
        }

        if eval:
            opts['eval'] = eval

        if not sync and allow_nested:
            rpc_sync: Union[bool, Literal['urgent']] = "urgent"
        else:
            rpc_sync = sync

        f._nvim_rpc_spec = {  # type: ignore[attr-defined]
            'type': 'autocmd',
            'name': name,
            'sync': rpc_sync,
            'opts': opts
        }
        return f
    return dec


def function(
    name: str,
    range: Union[bool, str, int] = False,
    sync: bool = False,
    allow_nested: bool = False,
    eval: Optional[str] = None
) -> Callable[[F], F]:
    """Tag a function or plugin method as a Nvim function handler."""
    def dec(f: F) -> F:
        f._nvim_rpc_method_name = (  # type: ignore[attr-defined]
            'function:{}'.format(name)
        )
        f._nvim_rpc_sync = sync  # type: ignore[attr-defined]
        f._nvim_bind = True  # type: ignore[attr-defined]
        f._nvim_prefix_plugin_path = True  # type: ignore[attr-defined]

        opts = {}

        if range:
            opts['range'] = '' if range is True else str(range)

        if eval:
            opts['eval'] = eval

        if not sync and allow_nested:
            rpc_sync: Union[bool, Literal['urgent']] = "urgent"
        else:
            rpc_sync = sync

        f._nvim_rpc_spec = {  # type: ignore[attr-defined]
            'type': 'function',
            'name': name,
            'sync': rpc_sync,
            'opts': opts
        }
        return f
    return dec


def shutdown_hook(f: F) -> F:
    """Tag a function or method as a shutdown hook."""
    f._nvim_shutdown_hook = True  # type: ignore[attr-defined]
    f._nvim_bind = True  # type: ignore[attr-defined]
    return f


def decode(mode: str = unicode_errors_default) -> Callable[[F], F]:
    """Configure automatic encoding/decoding of strings."""
    def dec(f: F) -> F:
        f._nvim_decode = mode  # type: ignore[attr-defined]
        return f
    return dec


def encoding(encoding: Union[bool, str] = True) -> Callable[[F], F]:
    """DEPRECATED: use pynvim.decode()."""
    if isinstance(encoding, str):
        encoding = True

    def dec(f: F) -> F:
        f._nvim_decode = encoding  # type: ignore[attr-defined]
        return f
    return dec



"""Specifies pynvim version."""
# pylint: disable=consider-using-f-string

from types import SimpleNamespace

# see also setup.py
VERSION = SimpleNamespace(major=0, minor=6, patch=0, prerelease="")

# e.g. "0.5.0", "0.5.0.dev0" (PEP-440)
__version__ = '{major}.{minor}.{patch}'.format(**vars(VERSION))

if VERSION.prerelease:
    __version__ += '.' + VERSION.prerelease



# type: ignore
"""Legacy python/python3-vim emulation."""
import io
import logging
import os
import sys
from importlib.machinery import PathFinder
from types import ModuleType

from pynvim.api import Nvim, walk
from pynvim.msgpack_rpc import ErrorResponse
from pynvim.plugin.decorators import plugin, rpc_export
from pynvim.util import format_exc_skip

__all__ = ('ScriptHost',)


logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warning,)


@plugin
class ScriptHost(object):

    """Provides an environment for running python plugins created for Vim."""

    def __init__(self, nvim):
        """Initialize the legacy python-vim environment."""
        self.setup(nvim)
        # context where all code will run
        self.module = ModuleType('__main__')
        nvim.script_context = self.module
        # it seems some plugins assume 'sys' is already imported, so do it now
        exec('import sys', self.module.__dict__)
        self.legacy_vim = LegacyVim.from_nvim(nvim)
        sys.modules['vim'] = self.legacy_vim
        # mimic Vim by importing vim module by default.
        exec('import vim', self.module.__dict__)
        # Handle DirChanged. #296
        nvim.command(
            'au DirChanged * call rpcnotify({}, "python_chdir", v:event.cwd)'
            .format(nvim.channel_id), async_=True)
        # XXX: Avoid race condition.
        # https://github.com/neovim/pynvim/pull/296#issuecomment-358970531
        # TODO(bfredl): when host initialization has been refactored,
        # to make __init__ safe again, the following should work:
        # os.chdir(nvim.eval('getcwd()', async_=False))
        nvim.command('call rpcnotify({}, "python_chdir", getcwd())'
                     .format(nvim.channel_id), async_=True)

    def setup(self, nvim):
        """Setup import hooks and global streams.

        This will add import hooks for importing modules from runtime
        directories and patch the sys module so 'print' calls will be
        forwarded to Nvim.
        """
        self.nvim = nvim
        info('install import hook/path')
        self.hook = path_hook(nvim)
        sys.path_hooks.append(self.hook)
        nvim.VIM_SPECIAL_PATH = '_vim_path_'
        sys.path.append(nvim.VIM_SPECIAL_PATH)
        info('redirect sys.stdout and sys.stderr')
        self.saved_stdout = sys.stdout
        self.saved_stderr = sys.stderr
        sys.stdout = RedirectStream(lambda data: nvim.out_write(data))
        sys.stderr = RedirectStream(lambda data: nvim.err_write(data))

    def teardown(self):
        """Restore state modified from the `setup` call."""
        nvim = self.nvim
        info('uninstall import hook/path')
        sys.path.remove(nvim.VIM_SPECIAL_PATH)
        sys.path_hooks.remove(self.hook)
        info('restore sys.stdout and sys.stderr')
        sys.stdout = self.saved_stdout
        sys.stderr = self.saved_stderr

    @rpc_export('python_execute', sync=True)
    def python_execute(self, script, range_start, range_stop):
        """Handle the `python` ex command."""
        self._set_current_range(range_start, range_stop)

        if script.startswith('='):
            # Handle ":py= ...". Evaluate as an expression and print.
            # (note: a valid python statement can't start with "=")
            expr = script[1:]
            print(self.python_eval(expr))
            return

        try:
            # pylint: disable-next=exec-used
            exec(script, self.module.__dict__)
        except Exception as exc:
            raise ErrorResponse(format_exc_skip(1)) from exc

    @rpc_export('python_execute_file', sync=True)
    def python_execute_file(self, file_path, range_start, range_stop):
        """Handle the `pyfile` ex command."""
        self._set_current_range(range_start, range_stop)
        with open(file_path, 'rb') as f:
            script = compile(f.read(), file_path, 'exec')
            try:
                # pylint: disable-next=exec-used
                exec(script, self.module.__dict__)
            except Exception as exc:
                raise ErrorResponse(format_exc_skip(1)) from exc

    @rpc_export('python_do_range', sync=True)
    def python_do_range(self, start, stop, code):
        """Handle the `pydo` ex command."""
        self._set_current_range(start, stop)
        nvim = self.nvim
        start -= 1
        fname = '_vim_pydo'

        # define the function
        function_def = 'def %s(line, linenr):\n %s' % (fname, code,)
        exec(function_def, self.module.__dict__)
        # get the function
        function = self.module.__dict__[fname]
        while start < stop:
            # Process batches of 5000 to avoid the overhead of making multiple
            # API calls for every line. Assuming an average line length of 100
            # bytes, approximately 488 kilobytes will be transferred per batch,
            # which can be done very quickly in a single API call.
            sstart = start
            sstop = min(start + 5000, stop)
            lines = nvim.current.buffer.api.get_lines(sstart, sstop, True)

            exception = None
            newlines = []
            linenr = sstart + 1
            for i, line in enumerate(lines):
                result = function(line, linenr)
                if result is None:
                    # Update earlier lines, and skip to the next
                    if newlines:
                        end = sstart + len(newlines) - 1
                        nvim.current.buffer.api.set_lines(sstart, end,
                                                          True, newlines)
                    sstart += len(newlines) + 1
                    newlines = []
                    pass
                elif isinstance(result, str):
                    newlines.append(result)
                else:
                    exception = TypeError('pydo should return a string '
                                          + 'or None, found %s instead'
                                          % result.__class__.__name__)
                    break
                linenr += 1

            start = sstop
            if newlines:
                end = sstart + len(newlines)
                nvim.current.buffer.api.set_lines(sstart, end, True, newlines)
            if exception:
                raise exception
        # delete the function
        del self.module.__dict__[fname]

    @rpc_export('python_eval', sync=True)
    def python_eval(self, expr):
        """Handle the `pyeval` vim function."""
        try:
            # pylint: disable-next=eval-used
            return eval(expr, self.module.__dict__)
        except Exception as exc:
            raise ErrorResponse(format_exc_skip(1)) from exc

    @rpc_export('python_chdir', sync=False)
    def python_chdir(self, cwd):
        """Handle working directory changes."""
        os.chdir(cwd)

    def _set_current_range(self, start, stop):
        current = self.legacy_vim.current
        current.range = current.buffer.range(start, stop)


class RedirectStream(io.IOBase):
    def __init__(self, redirect_handler):
        self.redirect_handler = redirect_handler

    def write(self, data):
        self.redirect_handler(data)

    def writelines(self, seq):
        self.redirect_handler('\n'.join(seq))


num_types = (int, float)


def num_to_str(obj):
    if isinstance(obj, num_types):
        return str(obj)
    else:
        return obj


class LegacyVim(Nvim):
    def eval(self, expr):
        obj = self.request("vim_eval", expr)
        return walk(num_to_str, obj)


# Copied/adapted from :help if_pyth.
def path_hook(nvim):
    def _get_paths():
        if nvim._thread_invalid():
            return []
        return discover_runtime_directories(nvim)

    def _find_module(fullname, oldtail, path):
        import imp
        idx = oldtail.find('.')
        if idx > 0:
            name = oldtail[:idx]
            tail = oldtail[idx + 1:]
            fmr = imp.find_module(name, path)
            module = imp.find_module(fullname[:-len(oldtail)] + name, *fmr)
            return _find_module(fullname, tail, module.__path__)
        else:
            return imp.find_module(fullname, path)

    class VimModuleLoader(object):
        def __init__(self, module):
            self.module = module

        def load_module(self, fullname, path=None):
            # Check sys.modules, required for reload (see PEP302).
            try:
                return sys.modules[fullname]
            except KeyError:
                pass
            import imp
            return imp.load_module(fullname, *self.module)

    class VimPathFinder(object):
        @staticmethod
        def find_module(fullname, path=None):
            """Method for Python 2.7 and 3.3."""
            try:
                return VimModuleLoader(
                    _find_module(fullname, fullname, path or _get_paths()))
            except ImportError:
                return None

        @staticmethod
        def find_spec(fullname, target=None):
            """Method for Python 3.4+."""
            return PathFinder.find_spec(fullname, _get_paths(), target)

    def hook(path):
        if path == nvim.VIM_SPECIAL_PATH:
            return VimPathFinder
        else:
            raise ImportError

    return hook


def discover_runtime_directories(nvim):
    rv = []
    for rtp in nvim.list_runtime_paths():
        if not os.path.exists(rtp):
            continue
        for subdir in ['pythonx', 'python3']:
            path = os.path.join(rtp, subdir)
            if os.path.exists(path):
                rv.append(path)
    return rv



"""Event loop abstraction subpackage.

We use python's built-in asyncio as the backend.
"""

from pynvim.msgpack_rpc.event_loop.asyncio import AsyncioEventLoop as EventLoop
from pynvim.msgpack_rpc.event_loop.base import TTransportType


__all__ = ['EventLoop', 'TTransportType']



"""Code for compatibility across Python versions."""
import warnings
from typing import Any, Dict, Optional


def find_module(fullname, path):  # type: ignore
    """Compatibility wrapper for imp.find_module.

    Automatically decodes arguments of find_module, in Python3
    they must be Unicode
    """
    if isinstance(fullname, bytes):
        fullname = fullname.decode()
    if isinstance(path, bytes):
        path = path.decode()
    elif isinstance(path, list):
        newpath = []
        for element in path:
            if isinstance(element, bytes):
                newpath.append(element.decode())
            else:
                newpath.append(element)
        path = newpath
    from imp import find_module as original_find_module
    return original_find_module(fullname, path)


unicode_errors_default = 'surrogateescape'

NUM_TYPES = (int, float)


def check_async(async_: Optional[bool], kwargs: Dict[str, Any], default: bool) -> bool:
    """Return a value of 'async' in kwargs or default when async_ is None.

    This helper function exists for backward compatibility (See #274).
    It shows a warning message when 'async' in kwargs is used to note users.
    """
    if async_ is not None:
        return async_
    elif 'async' in kwargs:
        warnings.warn(
            '"async" attribute is deprecated. Use "async_" instead.',
            DeprecationWarning,
        )
        return kwargs.pop('async')
    else:
        return default



"""Code shared between the API classes."""
import functools
import sys
from abc import ABC, abstractmethod
from typing import (Any, Callable, Generic, Iterator, List, Optional, Tuple, TypeVar,
                    Union, overload)

from msgpack import unpackb
if sys.version_info < (3, 8):
    from typing_extensions import Literal, Protocol
else:
    from typing import Literal, Protocol

from pynvim.compat import unicode_errors_default

__all__ = ()


T = TypeVar('T')
TDecodeMode = Union[Literal[True], str]


class NvimError(Exception):
    pass


class IRemote(Protocol):
    def request(self, name: str, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError


class Remote(ABC):

    """Base class for Nvim objects(buffer/window/tabpage).

    Each type of object has it's own specialized class with API wrappers around
    the msgpack-rpc session. This implements equality which takes the remote
    object handle into consideration.
    """

    def __init__(self, session: IRemote, code_data: Tuple[int, Any]):
        """Initialize from session and code_data immutable object.

        The `code_data` contains serialization information required for
        msgpack-rpc calls. It must be immutable for Buffer equality to work.
        """
        self._session = session
        self.code_data = code_data
        self.handle = unpackb(code_data[1])
        self.api = RemoteApi(self, self._api_prefix)
        self.vars = RemoteMap(self, self._api_prefix + 'get_var',
                              self._api_prefix + 'set_var',
                              self._api_prefix + 'del_var')
        self.options = RemoteMap(self, self._api_prefix + 'get_option',
                                 self._api_prefix + 'set_option')

    @property
    @abstractmethod
    def _api_prefix(self) -> str:
        raise NotImplementedError()

    def __repr__(self) -> str:
        """Get text representation of the object."""
        return '<%s(handle=%r)>' % (
            self.__class__.__name__,
            self.handle,
        )

    def __eq__(self, other: Any) -> bool:
        """Return True if `self` and `other` are the same object."""
        return (hasattr(other, 'code_data')
                and other.code_data == self.code_data)

    def __hash__(self) -> int:
        """Return hash based on remote object id."""
        return self.code_data.__hash__()

    def request(self, name: str, *args: Any, **kwargs: Any) -> Any:
        """Wrapper for nvim.request."""
        return self._session.request(name, self, *args, **kwargs)


class RemoteApi:
    """Wrapper to allow api methods to be called like python methods."""

    def __init__(self, obj: IRemote, api_prefix: str):
        """Initialize a RemoteApi with object and api prefix."""
        self._obj = obj
        self._api_prefix = api_prefix

    def __getattr__(self, name: str) -> Callable[..., Any]:
        """Return wrapper to named api method."""
        return functools.partial(self._obj.request, self._api_prefix + name)


E = TypeVar('E', bound=Exception)


def transform_keyerror(exc: E) -> Union[E, KeyError]:
    if isinstance(exc, NvimError):
        if exc.args[0].startswith('Key not found:'):
            return KeyError(exc.args[0])
        if exc.args[0].startswith('Invalid option name:'):
            return KeyError(exc.args[0])
    return exc


class RemoteMap:
    """Represents a string->object map stored in Nvim.

    This is the dict counterpart to the `RemoteSequence` class, but it is used
    as a generic way of retrieving values from the various map-like data
    structures present in Nvim.

    It is used to provide a dict-like API to vim variables and options.
    """

    _set = None
    _del = None

    def __init__(
        self,
        obj: IRemote,
        get_method: str,
        set_method: Optional[str] = None,
        del_method: Optional[str] = None
    ):
        """Initialize a RemoteMap with session, getter/setter."""
        self._get = functools.partial(obj.request, get_method)
        if set_method:
            self._set = functools.partial(obj.request, set_method)
        if del_method:
            self._del = functools.partial(obj.request, del_method)

    def __getitem__(self, key: str) -> Any:
        """Return a map value by key."""
        try:
            return self._get(key)
        except NvimError as exc:
            raise transform_keyerror(exc)

    def __setitem__(self, key: str, value: Any) -> None:
        """Set a map value by key(if the setter was provided)."""
        if not self._set:
            raise TypeError('This dict is read-only')
        self._set(key, value)

    def __delitem__(self, key: str) -> None:
        """Delete a map value by associating None with the key."""
        if not self._del:
            raise TypeError('This dict is read-only')
        try:
            return self._del(key)
        except NvimError as exc:
            raise transform_keyerror(exc)

    def __contains__(self, key: str) -> bool:
        """Check if key is present in the map."""
        try:
            self._get(key)
            return True
        except Exception:
            return False

    @overload
    def get(self, key: str, default: T) -> T: ...

    @overload
    def get(self, key: str, default: Optional[T] = None) -> Optional[T]: ...

    def get(self, key: str, default: Optional[T] = None) -> Optional[T]:
        """Return value for key if present, else a default value."""
        try:
            return self.__getitem__(key)
        except KeyError:
            return default


class RemoteSequence(Generic[T]):

    """Represents a sequence of objects stored in Nvim.

    This class is used to wrap msgpack-rpc functions that work on Nvim
    sequences(of lines, buffers, windows and tabpages) with an API that
    is similar to the one provided by the python-vim interface.

    For example, the 'windows' property of the `Nvim` class is a RemoteSequence
    sequence instance, and the expression `nvim.windows[0]` is translated to
    session.request('nvim_list_wins')[0].

    One important detail about this class is that all methods will fetch the
    sequence into a list and perform the necessary manipulation
    locally(iteration, indexing, counting, etc).
    """

    def __init__(self, session: IRemote, method: str):
        """Initialize a RemoteSequence with session, method."""
        self._fetch = functools.partial(session.request, method)

    def __len__(self) -> int:
        """Return the length of the remote sequence."""
        return len(self._fetch())

    @overload
    def __getitem__(self, idx: int) -> T: ...

    @overload
    def __getitem__(self, idx: slice) -> List[T]: ...

    def __getitem__(self, idx: Union[slice, int]) -> Union[T, List[T]]:
        """Return a sequence item by index."""
        if not isinstance(idx, slice):
            return self._fetch()[idx]
        return self._fetch()[idx.start:idx.stop]

    def __iter__(self) -> Iterator[T]:
        """Return an iterator for the sequence."""
        items = self._fetch()
        for item in items:
            yield item

    def __contains__(self, item: T) -> bool:
        """Check if an item is present in the sequence."""
        return item in self._fetch()


@overload
def decode_if_bytes(obj: bytes, mode: TDecodeMode = True) -> str: ...


@overload
def decode_if_bytes(obj: T, mode: TDecodeMode = True) -> Union[T, str]: ...


def decode_if_bytes(obj: T, mode: TDecodeMode = True) -> Union[T, str]:
    """Decode obj if it is bytes."""
    if mode is True:
        mode = unicode_errors_default
    if isinstance(obj, bytes):
        return obj.decode("utf-8", errors=mode)
    return obj


def walk(fn: Callable[[Any], Any], obj: Any) -> Any:
    """Recursively walk an object graph applying `fn` to objects."""

    # Note: this function is very hot, so it is worth being careful
    # about performance.
    type_ = type(obj)

    if type_ is list or type_ is tuple:
        return [walk(fn, o) for o in obj]
    if type_ is dict:
        return {walk(fn, k): walk(fn, v) for k, v in obj.items()}
    return fn(obj)



"""Synchronous msgpack-rpc session layer."""
import logging
import sys
import threading
from collections import deque
from traceback import format_exc
from typing import (Any, AnyStr, Callable, Deque, List, NamedTuple, Optional, Sequence,
                    Tuple, Union, cast)

import greenlet

from pynvim.compat import check_async
from pynvim.msgpack_rpc.async_session import AsyncSession
from pynvim.msgpack_rpc.event_loop.base import BaseEventLoop

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

logger = logging.getLogger(__name__)
error, debug, info, warn = (logger.error, logger.debug, logger.info,
                            logger.warning,)


class Request(NamedTuple):
    """A request from Nvim."""

    type: Literal['request']
    name: str
    args: List[Any]
    response: Any


class Notification(NamedTuple):
    """A notification from Nvim."""

    type: Literal['notification']
    name: str
    args: List[Any]


Message = Union[Request, Notification]


class Session:

    """Msgpack-rpc session layer that uses coroutines for a synchronous API.

    This class provides the public msgpack-rpc API required by this library.
    It uses the greenlet module to handle requests and notifications coming
    from Nvim with a synchronous API.
    """

    def __init__(self, async_session: AsyncSession):
        """Wrap `async_session` on a synchronous msgpack-rpc interface."""
        self._async_session = async_session
        self._request_cb: Optional[Callable[[str, List[Any]], None]] = None
        self._notification_cb: Optional[Callable[[str, List[Any]], None]] = None
        self._pending_messages: Deque[Message] = deque()
        self._is_running = False
        self._setup_exception: Optional[Exception] = None
        self._loop_thread: Optional[threading.Thread] = None
        self.error_wrapper: Callable[[Tuple[int, str]], Exception] = \
            lambda e: Exception(e[1])

    @property
    def loop(self) -> BaseEventLoop:
        """Get the underlying msgpack EventLoop."""
        return self._async_session.loop

    def threadsafe_call(
        self, fn: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> None:
        """Wrapper around `AsyncSession.threadsafe_call`."""
        def handler():
            try:
                fn(*args, **kwargs)
            except Exception:
                warn("error caught while executing async callback\n%s\n",
                     format_exc())

        def greenlet_wrapper():
            gr = greenlet.greenlet(handler)
            gr.switch()

        self._async_session.threadsafe_call(greenlet_wrapper)

    def next_message(self) -> Optional[Message]:
        """Block until a message(request or notification) is available.

        If any messages were previously enqueued, return the first in queue.
        If not, run the event loop until one is received.
        """
        if self._is_running:
            raise Exception('Event loop already running')
        if self._pending_messages:
            return self._pending_messages.popleft()
        self._async_session.run(self._enqueue_request_and_stop,
                                self._enqueue_notification_and_stop)
        if self._pending_messages:
            return self._pending_messages.popleft()
        return None

    def request(self, method: AnyStr, *args: Any, **kwargs: Any) -> Any:
        """Send a msgpack-rpc request and block until as response is received.

        If the event loop is running, this method must have been called by a
        request or notification handler running on a greenlet. In that case,
        send the quest and yield to the parent greenlet until a response is
        available.

        When the event loop is not running, it will perform a blocking request
        like this:
        - Send the request
        - Run the loop until the response is available
        - Put requests/notifications received while waiting into a queue

        If the `async_` flag is present and True, a asynchronous notification
        is sent instead. This will never block, and the return value or error
        is ignored.
        """
        async_ = check_async(kwargs.pop('async_', None), kwargs, False)
        if async_:
            self._async_session.notify(method, args)
            return

        if kwargs:
            raise ValueError("request got unsupported keyword argument(s): {}"
                             .format(', '.join(kwargs.keys())))

        if self._is_running:
            v = self._yielding_request(method, args)
        else:
            v = self._blocking_request(method, args)
        if not v:
            # EOF
            raise OSError('EOF')
        err, rv = v
        if err:
            info("'Received error: %s", err)
            raise self.error_wrapper(err)
        return rv

    def run(self,
            request_cb: Callable[[str, List[Any]], None],
            notification_cb: Callable[[str, List[Any]], None],
            setup_cb: Optional[Callable[[], None]] = None) -> None:
        """Run the event loop to receive requests and notifications from Nvim.

        Like `AsyncSession.run()`, but `request_cb` and `notification_cb` are
        inside greenlets.
        """
        self._request_cb = request_cb
        self._notification_cb = notification_cb
        self._is_running = True
        self._setup_exception = None
        self._loop_thread = threading.current_thread()

        def on_setup() -> None:
            try:
                setup_cb()  # type: ignore[misc]
            except Exception as e:
                self._setup_exception = e
                self.stop()

        if setup_cb:
            # Create a new greenlet to handle the setup function
            gr = greenlet.greenlet(on_setup)
            gr.switch()

        if self._setup_exception:
            error(  # type: ignore[unreachable]
                'Setup error: {}'.format(self._setup_exception)
            )
            raise self._setup_exception

        # Process all pending requests and notifications
        while self._pending_messages:
            msg = self._pending_messages.popleft()
            getattr(self, '_on_{}'.format(msg[0]))(*msg[1:])
        self._async_session.run(self._on_request, self._on_notification)
        self._is_running = False
        self._request_cb = None
        self._notification_cb = None
        self._loop_thread = None

        if self._setup_exception:
            raise self._setup_exception

    def stop(self) -> None:
        """Stop the event loop."""
        self._async_session.stop()

    def close(self) -> None:
        """Close the event loop."""
        self._async_session.close()

    def _yielding_request(
        self, method: AnyStr, args: Sequence[Any]
    ) -> Tuple[Tuple[int, str], Any]:
        gr = greenlet.getcurrent()
        parent = gr.parent

        def response_cb(err, rv):
            debug('response is available for greenlet %s, switching back', gr)
            gr.switch(err, rv)

        self._async_session.request(method, args, response_cb)
        debug('yielding from greenlet %s to wait for response', gr)
        return parent.switch()

    def _blocking_request(
            self, method: AnyStr, args: Sequence[Any]
    ) -> Tuple[Tuple[int, str], Any]:
        result = []

        def response_cb(err, rv):
            result.extend([err, rv])
            self.stop()

        self._async_session.request(method, args, response_cb)
        self._async_session.run(self._enqueue_request,
                                self._enqueue_notification)
        return cast(Tuple[Tuple[int, str], Any], tuple(result))

    def _enqueue_request_and_stop(
        self, name: str, args: List[Any], response: Any
    ) -> None:
        self._enqueue_request(name, args, response)
        self.stop()

    def _enqueue_notification_and_stop(self, name: str, args: List[Any]) -> None:
        self._enqueue_notification(name, args)
        self.stop()

    def _enqueue_request(self, name: str, args: List[Any], response: Any) -> None:
        self._pending_messages.append(Request('request', name, args, response,))

    def _enqueue_notification(self, name: str, args: List[Any]) -> None:
        self._pending_messages.append(Notification('notification', name, args,))

    def _on_request(self, name, args, response):
        def handler():
            try:
                rv = self._request_cb(name, args)
                debug('greenlet %s finished executing, '
                      + 'sending %s as response', gr, rv)
                response.send(rv)
            except ErrorResponse as err:
                debug("error response from request '%s %s': %s",
                      name, args, format_exc())
                response.send(err.args[0], error=True)
            except Exception as err:
                warn("error caught while processing request '%s %s': %s",
                     name, args, format_exc())
                response.send(repr(err) + "\n" + format_exc(5), error=True)
            debug('greenlet %s is now dying...', gr)

        # Create a new greenlet to handle the request
        gr = greenlet.greenlet(handler)
        debug('received rpc request, greenlet %s will handle it', gr)
        gr.switch()

    def _on_notification(self, name, args):
        def handler():
            try:
                self._notification_cb(name, args)
                debug('greenlet %s finished executing', gr)
            except Exception:
                warn("error caught while processing notification '%s %s': %s",
                     name, args, format_exc())

            debug('greenlet %s is now dying...', gr)

        gr = greenlet.greenlet(handler)
        debug('received rpc notification, greenlet %s will handle it', gr)
        gr.switch()


class ErrorResponse(BaseException):

    """Raise this in a request handler to respond with a given error message.

    Unlike when other exceptions are caught, this gives full control off the
    error response sent. When "ErrorResponse(msg)" is caught "msg" will be
    sent verbatim as the error response.No traceback will be appended.
    """

    pass



"""Python client for Nvim.

Client library for talking with Nvim processes via its msgpack-rpc API.
"""
import logging
import os
import sys
from types import SimpleNamespace as Version
from typing import List, Optional, cast, overload

from pynvim._version import VERSION, __version__
from pynvim.api import Nvim, NvimError
from pynvim.msgpack_rpc import (ErrorResponse, Session, TTransportType,
                                child_session, socket_session, stdio_session,
                                tcp_session)
from pynvim.plugin import (Host, autocmd, command, decode, encoding, function,
                           plugin, rpc_export, shutdown_hook)

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal


__all__ = ('tcp_session', 'socket_session', 'stdio_session', 'child_session',
           'start_host', 'autocmd', 'command', 'encoding', 'decode',
           'function', 'plugin', 'rpc_export', 'Host', 'Nvim', 'NvimError',
           'Version', 'VERSION', '__version__',
           'shutdown_hook', 'attach', 'setup_logging', 'ErrorResponse',
           )


def start_host(session: Optional[Session] = None) -> None:
    """Promote the current process into python plugin host for Nvim.

    Start msgpack-rpc event loop for `session`, listening for Nvim requests
    and notifications. It registers Nvim commands for loading/unloading
    python plugins.

    The sys.stdout and sys.stderr streams are redirected to Nvim through
    `session`. That means print statements probably won't work as expected
    while this function doesn't return.

    This function is normally called at program startup and could have been
    defined as a separate executable. It is exposed as a library function for
    testing purposes only.
    """
    plugins = []
    for arg in sys.argv:
        _, ext = os.path.splitext(arg)
        if ext == '.py':
            plugins.append(arg)
        elif os.path.isdir(arg):
            init = os.path.join(arg, '__init__.py')
            if os.path.isfile(init):
                plugins.append(arg)

    # This is a special case to support the old workaround of
    # adding an empty .py file to make a package directory
    # visible, and it should be removed soon.
    for path in list(plugins):
        dup = path + ".py"
        if os.path.isdir(path) and dup in plugins:
            plugins.remove(dup)

    # Special case: the legacy scripthost receives a single relative filename
    # while the rplugin host will receive absolute paths.
    if plugins == ["script_host.py"]:
        name = "script"
    else:
        name = "rplugin"

    setup_logging(name)

    if not session:
        session = stdio_session()
    nvim = Nvim.from_session(session)

    if nvim.version.api_level < 1:
        sys.stderr.write("This version of pynvim "
                         "requires nvim 0.1.6 or later")
        sys.exit(1)

    host = Host(nvim)
    host.start(plugins)


@overload
def attach(session_type: Literal['tcp'], address: str, port: int = 7450) -> Nvim: ...


@overload
def attach(session_type: Literal['socket'], *, path: str) -> Nvim: ...


@overload
def attach(session_type: Literal['child'], *, argv: List[str]) -> Nvim: ...


@overload
def attach(session_type: Literal['stdio']) -> Nvim: ...


def attach(
    session_type: TTransportType,
    address: Optional[str] = None,
    port: int = 7450,
    path: Optional[str] = None,
    argv: Optional[List[str]] = None,
    decode: Literal[True] = True
) -> Nvim:
    """Provide a nicer interface to create python api sessions.

    Previous machinery to create python api sessions is still there. This only
    creates a facade function to make things easier for the most usual cases.
    Thus, instead of:
        from pynvim import socket_session, Nvim
        session = tcp_session(address=<address>, port=<port>)
        nvim = Nvim.from_session(session)
    You can now do:
        from pynvim import attach
        nvim = attach('tcp', address=<address>, port=<port>)
    And also:
        nvim = attach('socket', path=<path>)
        nvim = attach('child', argv=<argv>)
        nvim = attach('stdio')

    When the session is not needed anymore, it is recommended to explicitly
    close it:
       nvim.close()
    It is also possible to use the session as a context manager:
       with attach('socket', path=thepath) as nvim:
           print(nvim.funcs.getpid())
           print(nvim.current.line)
    This will automatically close the session when you're done with it, or
    when an error occurred.


    """
    session = (
        tcp_session(cast(str, address), port) if session_type == 'tcp' else
        socket_session(cast(str, path)) if session_type == 'socket' else
        stdio_session() if session_type == 'stdio' else
        child_session(cast(List[str], argv)) if session_type == 'child' else
        None
    )

    if not session:
        raise Exception('Unknown session type "%s"' % session_type)

    return Nvim.from_session(session).with_decode(decode)


def setup_logging(name: str) -> None:
    """Setup logging according to environment variables."""
    logger = logging.getLogger(__name__)
    if 'NVIM_PYTHON_LOG_FILE' in os.environ:
        prefix = os.environ['NVIM_PYTHON_LOG_FILE'].strip()
        major_version = sys.version_info[0]
        logfile = '{}_py{}_{}'.format(prefix, major_version, name)
        handler = logging.FileHandler(logfile, 'w', 'utf-8')
        handler.formatter = logging.Formatter(
            '%(asctime)s [%(levelname)s @ '
            '%(filename)s:%(funcName)s:%(lineno)s] %(process)s - %(message)s')
        logging.root.addHandler(handler)
        level = logging.INFO
        env_log_level = os.environ.get('NVIM_PYTHON_LOG_LEVEL', None)
        if env_log_level is not None:
            lvl = getattr(logging, env_log_level.strip(), None)
            if isinstance(lvl, int):
                level = lvl
            else:
                logger.warning('Invalid NVIM_PYTHON_LOG_LEVEL: %r, using INFO.',
                               env_log_level)
        logger.setLevel(level)
